# from pathlib import Path

# from dotenv import load_dotenv

# # Load environment variables from .env file if it exists
# load_dotenv()

# # Paths
# PROJ_ROOT = Path(__file__).resolve().parents[1]
# # logger.info(f"PROJ_ROOT path is: {PROJ_ROOT}")

# DATA_DIR = PROJ_ROOT / "data"
# MODELS_DIR = PROJ_ROOT / "models"
# REPORTS_DIR = PROJ_ROOT / "reports"
# FIGURES_DIR = REPORTS_DIR / "figures"

# # If tqdm is installed, configure loguru with tqdm.write
# # https://github.com/Delgan/loguru/issues/135
# # try:
# #     from tqdm import tqdm

# #     logger.remove(0)
# #     logger.add(lambda msg: tqdm.write(msg, end=""), colorize=True)
# # except ModuleNotFoundError:
# #     pass

# Not used atm


from .openai import ChatGPT
import numpy as np
import concurrent
class ParallelizedModel:
    def __init__(self, model, n_threads):
        self.model = model
        self.n_threads = n_threads

    def single(self, *args):
        single_input, input_index = args[0]
        prediction = self.model(single_input)
        self.results[input_index] = prediction

    def __call__(self, input_list):
        """
        For the list of inpiut call the model and return all the string responses.
        """
        self.results = [None] * len(input_list)
        samples = [(input_list[i], i) for i in range(len(input_list))]
        with concurrent.futures.ThreadPoolExecutor(max_workers=self.n_threads) as executor:
            futures = {executor.submit(self.single, sample): sample for sample in samples}
            for future in concurrent.futures.as_completed(futures):
                future.result()
                del futures[future]  # Crucial to close memory leaks
        return self.results

    def set_template(self, template):
        self.model.set_template(template)

    def get_template(self):
        return self.model.template

def load_model(model_name, **kwargs):
    match model_name.lower():
        case "chatgpt":
            n_responses = kwargs["n_responses"] if "n_responses" in kwargs else 8
            n_threads = kwargs["n_threads"] if "n_threads" in kwargs else 4
            temperature = kwargs["temperature"] if "temperature" in kwargs else 1.0
            max_tokens = kwargs["max_tokens"] if "max_tokens" in kwargs else 250
            cache_dir = kwargs["cache_dir"] if "cache_dir" in kwargs else None
            return ParallelizedModel(
                ChatGPT(n_responses=n_responses, max_tokens=max_tokens, temperature=temperature, cache_dir=cache_dir),
                n_threads,
            )
        case _ : raise ValueError(f"Model {model_name} not supported")



from transformers import AutoTokenizer, AutoModelForSeq2SeqLM
import torch
import torch.nn as nn
import os
import numpy as np
from dotenv import load_dotenv

load_dotenv()

cache_dir = os.getenv("HUGGINGFACE_CACHE")

allowed_models = [
    "google/flan-t5-large",
]
# Stores which tokenizer to use for which model
model_name_to_tokenizer = {
    "google/flan-t5-large": "google/flan-t5-large",
}
model_name_to_class = {
    "google/flan-t5-large": AutoModelForSeq2SeqLM,
}


class GenerativeModel:
    def __init__(self, model_name: str):
        self.tokenizer = AutoTokenizer.from_pretrained(
            model_name_to_tokenizer[model_name], cache_dir=cache_dir
        )
        self.model = model_name_to_class[model_name].from_pretrained(
            model_name, cache_dir=cache_dir, device_map="auto"
        )

    def generate(self, *args):
        return self.model.generate(*args)

    def get_logprobability(self, input_ids, target_ids):
        """Get the log probability of the target sequence given the input sequence."""
        loss_fct = nn.NLLLoss(reduction="none")
        lsm = nn.LogSoftmax(dim=1)  # first applies the softmax to ensure all values are normalized
        self.model.eval()
        with torch.no_grad():
            output = self.model(
                input_ids=input_ids,
                labels=target_ids,
            )
        # print("Output logits_shape: ", output.logits.shape)
        logits = output.logits
        lsm_logits = lsm(logits)
        # Gather the log probabilities associated with the token ids given by target_ids
        log_probs = lsm_logits.gather(2, target_ids.unsqueeze(-1)).squeeze(-1)
        summed_log_probs = log_probs.sum(1)
        # Convert to numpy array for SHAP
        summed_log_probs = summed_log_probs.to("cpu").numpy()
        return summed_log_probs

    def get_tokens_and_groups_from_continuous_spans(self, spans: list[str], add_eos=False):
        input_tokens = []
        span_indices = []
        current_index = 0
        for span in spans:
            sent_tokens = self.tokenizer(
                span, return_tensors="pt", add_special_tokens=False
            ).input_ids[0]
            # add_special_tokens=False to avoid adding eos token
            input_tokens.extend(sent_tokens)
            span_indices.extend([current_index] * len(sent_tokens))
            current_index += 1

        if add_eos:
            input_tokens.append(self.tokenizer.eos_token_id)
            span_indices.append(current_index - 1)
        return torch.stack(input_tokens), span_indices

    def tokenize(self, text: str):
        return self.tokenizer(text, return_tensors="pt", add_special_tokens=False).input_ids[0]

    def decode(self, tokens):
        return self.tokenizer.decode(tokens)



import enum
import numpy as np
import pandas as pd
from shap import KernelExplainer #, PartitionExplainer
# from shap.maskers import Masker
import matplotlib.pyplot as plt
import os
from os.path import join as pjoin

from . import util
from .models import load_model
from .containers import AssignedInput, ExplainedInput, EvaluatedExplanation


class Operator(enum.Enum):
    START_WITH = "START_WITH"
    END_WITH = "END_WITH"
    EQUAL = "EQUAL"
    CONTAIN = "CONTAIN"
    # The following operators are not supported in this version
    # SEMANTIC_EQUAL = "SEMANTIC_EQUAL"
    # ENTAIL = "ENTAIL"
    # CONTRADICT = "CONTRADICT"


class CAFGA:
    def __init__(self, model, model_params=None, explainer="KernelSHAP", explainer_params=None):
        """
        Create a CAFGA object.

        params
        ------
        model: str | callable
            The model whose predictions are to be explained. The model should be a callable that takes a list of strings as input and either returns list containing an output for each input. 
            The output can be of two kinds:
            1. The output can be a list of floats each representing a separate prediction on the same input (e.g. sentiment classifier).
            2. As an intermediate step the output may also be a list of string responses from the model (e.g. generative model)

            In case 2. the output still needs to be post processed into a list of float predictions. This can be done by providing a scalarizer to the explain function.
            The explain function will generate attributions for each float prediction separately. 
            To use a predefined model you can pass a string with the name of the model. The available option is "chatgpt".

        model_params: dict
            Parameters for the predefined model. (Not used if a custom model is provided)
        explainer: str
            The explainer method to be used. Currently only KernelSHAP is supported.
        """
        if type(model) == str:
            if model_params is None:
                model_params = {}
            self.model = load_model(model, **model_params)
        else:
            self.model = model

    def get_original_predicitons(self, input_segments : list[str]):
        input = "".join(input_segments)
        return self.f([input])

    def mask_to_input(self, mask):
        input = ""
        prev_was_masked = False
        for i in range(self.input_length):
            # If mask is False, then the input is replaced with the mask value for that input
            # Else the input is included as is
            if not mask[self.group_assignments[i]]:
                if not prev_was_masked or (not self.merge_masks):
                    # Place the mask value if 1. not merging masks or 2. merging masks but the previous input was not masked
                    input += self.mask_value[i]
                prev_was_masked = True
            else:
                input += self.input[i]
                prev_was_masked = False

        return input

    def mask_to_output(self, mask_list):
        input_list = [self.mask_to_input(mask) for mask in mask_list]
        return self.f(input_list)

    def init_scalarizer(self, scalarizer):
        if type(scalarizer) == list:
            # the model returns a string and the scalarizer is from the predefined list
            def predefined_scalarizer(responses):
                results = []
                for operator, target in scalarizer:
                    results.append(util.evaluate_outcome(responses, operator, target))
                return results

            self.scalarizer = predefined_scalarizer
        else:
            self.scalarizer = scalarizer
        self.uses_scalarizer = self.scalarizer is not None
        if self.uses_scalarizer:
            dummy_output = self.scalarizer([""])
        else:
            dummy_output = self.model([""])[0]
        if hasattr(dummy_output, "__len__") :
            self.scalarizer_output_dim = len(dummy_output)
        else:
            self.scalarizer_output_dim = 1

    def _explain(
        self,
        input: list[str],
        group_assignments: list[int],
        mask_value: str | list[str],
        merge_masks=False,
        partition_tree=None,
        n_allowed_samples=512,
        track_execution_data=False,
    ):
        """Uses the KernelExplainer from SHAP to explain the model's predictions on the input groups.

        Parameters
        ----------
        input : The input to the model in the form of a list of strings or token_ids (as pytorch tensor).

        group_assignments : A list of integers representing the group assignments of the input, where group_assignments[i] is the group of input[i].

        mask_value : The value (string or token) to use to mask the input when the group is removed.
        If provided with a list of strings, the input will be masked with the corresponding string.
        E.g. the i-th input will be masked with mask_value[i] (thus len(mask_value) should be equal to len(input)).

        merge_masks : Whether to merge consecutive masks into a single mask. Default is False.

        track_execution_data : Whether to track the time taken and number of samples generated during the computation. Default is False.

        Returns
        -------
        shap_values : The SHAP values for the input groups.
        If track_execution_data is True, it also returns:
        total_time : The time taken to compute the SHAP values.
        numbers_of_samples_generated : The number of samples generated during the computation. (<= n_allowed_samples+2)

        """
        self.input = input
        self.input_length = len(input)
        self.group_assignments = util.standardize_groups(group_assignments)
        if type(mask_value) != list:
            # Use a list in all cases to simplify the masking process
            mask_value = [mask_value] * self.input_length
        self.mask_value = mask_value
        self.merge_masks = merge_masks
        self.groups = set(self.group_assignments)
        self.n_groups = len(self.groups)
        self.track_execution_data = track_execution_data

        if track_execution_data:
            import time
            self.numbers_of_samples_generated = 0
            start = time.time()

        if partition_tree is not None:
            raise NotImplementedError("Using the partition tree is not supported in this version.")
            # Received a partition tree -> use PartitionExplainer on top of given groups
            masker = self.get_masker_for_partitionshap(partition_tree)
            sub_explainer = PartitionExplainer(
                self.f, masker
            )
            # All that partition shap needs is the number of inputs and the partition tree
            # The actual input is produced by the masker.
            # Hence, we pass in a dummy input whose only purpose is to declare the number of inputs

            self.shap_values = sub_explainer(["dummy"]).values[0]
        else:
            # No partition tree -> use KernelExplainer on the groups
            empty_mask = pd.Series([False] * self.n_groups)
            full_mask = pd.Series([True] * self.n_groups)
            self.shap_values = KernelExplainer(self.mask_to_output, empty_mask).shap_values(
                full_mask, 
                nsamples=n_allowed_samples
            )
        if track_execution_data:
            end = time.time()
            total_time = end - start
            return self.shap_values, total_time, self.numbers_of_samples_generated

        return self.shap_values

    def explain(
        self,
        input: str = None,
        assignment_method: str = None,
        segmented_input : list[str] = None,
        custom_assignments : list[int] = None,
        mask_value : str = "",
        merge_masks : bool = True,
        scalarizer = None,
        template : str = None,
        n_allowed_samples = 256,
    ):
        """
        Explain the model's prediction on the given input under the group assignment method proivded.

        params
        ------
        input: str 
            The unsegmented input to the model. If segmented_input is provided, this is ignored. The input will be segmented according to the assignment_method provided.
        assignment_method: str 
            The name of a predefined assignment method. The options are "word", "sentence" and "paragraph". ("syntax-tree" is not available in this version).
        segmented_input: list[str]
            The segments of the input to be explained. Custom assignments must be provided alongside this. This can be used to run the explanation on a custom assignment method. 
        custom_assignments: list[int]
            The custom group assignments for the segmented_input. This must be provided alongside segmented_input. The assignments should be formatted such that custom_assignments[i] is the group that segmented_input[i] belongs to.
        mask_value: str
            The value to use to mask the input when the group is removed.
        merge_masks: bool
            Whether to merge consecutive masks into a single mask.
        scalarizer:
            The scalarizer to use to convert the model's output (string) to a list of floats (each representing a prediction derived from the model's reponse). If None, it is assumed that the model provided at initialization returns a float as output.
            To use a one of the provided scalarizers pass a list of tuples [(operator, target)] where operator is one of the predefined operators and target is the target value to be used in the operator.
        template: str
            If you have a template that the input should be pasted into before being passed to the model, provide it here.
            This assumes that the model you provided has a set_template method that takes a string as input.
        n_allowed_samples: int
            The number of samples to generate for the KernelExplainer. Default is 256.

        Returns
        -------
        ExplainedInput: An object containing the input, the group assignments, the attributions, the mask_value, and the merge_masks value.
        Note that attributions is a list of lists of the shape: [n_scalar_outputs, n_groups] where n_scalar_outputs is the number of scalar outputs of the scalarizer/ the original model.

        """
        if segmented_input is None or custom_assignments is None:
            # parse the input
            match assignment_method:
                case "word":
                    input_segments = util.word_tokenize(input)
                    assignments = range(len(input_segments))
                case "sentence":
                    input_segments = util.sentence_tokenize(input)
                    assignments = range(len(input_segments))
                case "paragraph":
                    input_segments = util.paragraph_tokenize(input)
                    assignments = range(len(input_segments))
                case "syntax-tree":
                    input_segments, assignments = util.syntax_tokenize(input)
        else:
            input_segments = segmented_input
            assignments = custom_assignments

        self.init_scalarizer(scalarizer)

        def f(input_list):
            model_predictions = self.model(input_list)
            if self.uses_scalarizer:
                scalarized_predictions = np.zeros((len(input_list), self.scalarizer_output_dim))
                for i in range(len(input_list)):
                    scalarized_predictions[i] = self.scalarizer(model_predictions[i])
                return scalarized_predictions
            else:
                return model_predictions

        self.f = f
        if template is not None:
            try:
                self.model.set_template(template)
            except Exception as e:
                print("Encountered the following error when trying to set the template:")
                print(e)
                print("If you would like to use a template, your model should have a function: \"set_template\"")
        explanation = self._explain(input_segments, assignments, mask_value, merge_masks, n_allowed_samples=n_allowed_samples)
        if self.scalarizer_output_dim > 1:
            # Transpose from (n_groups, n_outputs) to (n_outputs, n_groups)
            explanation = [[attribution[d] for attribution in explanation] for d in range(self.scalarizer_output_dim)]
        else:
            # Only have one output for which we generate attributions
            explanation = [explanation] 
        explained_input = ExplainedInput(
            input_segments, template, assignments, explanation, mask_value, merge_masks
        )
        self.most_recent_explanation = explained_input
        return explained_input

    def compute_difference_original_to_perturbed(
        self, original_prediction, input_perturbations, scalarizer_index, difference_metric=None
    ):
        """Compute the difference of the explanation on the perturbed inputs as the difference between the original prediction and the perturbed prediction."""
        if difference_metric is None:
            difference_metric = util.l1

        differences = np.zeros(len(input_perturbations))
        perturbed_predictions = self.f(input_perturbations)
        for j, perturbed_prediction in enumerate(perturbed_predictions):
            difference = difference_metric(
                original_prediction[0][scalarizer_index], perturbed_prediction[scalarizer_index]
            )
            differences[j] = difference
        return differences

    def evaluate(
        self,
        explained_input: ExplainedInput,
        scalarizer,
        direction : str,
        step_size = 1,
        cut_off = 1.0,
        is_descedning=True,
        flip_signs=False,
    ):
        """
        Evaluate the explanation provided by the explain function. 
        As the two forms of evaluation are deletion and insertion, evaluating means calculating the differences between the original prediction and the prediction on the perturbed inputs.

        Params
        ------
        explained_input: ExplainedInput 
            The explanation to evaluate. Recevied from running the explain function.
        scalarizer:
            The same scalarizer used in the explain function. 
            (Not stored in the ExplainedInput object because the scalarizer may be a function and not serializable)
        direction: str
            The direction of perturbation. All ones to all zeroes (deletion) or all zeroes to all ones (insertion).

        Returns
        -------
        evaluated_explanation: EvaluatedExplanation
          An object containing the explained input, the parameters of the evaluation and the resulting differences.
        """
        self.init_scalarizer(scalarizer)
        def f(input_list):
            model_predictions = self.model(input_list)
            if self.uses_scalarizer:
                scalarized_predictions = np.zeros((len(input_list), self.scalarizer_output_dim))
                for i in range(len(input_list)):
                    scalarized_predictions[i] = self.scalarizer(model_predictions[i])
                return scalarized_predictions
            else:
                return model_predictions

        if explained_input.template is not None:
            self.model.set_template(explained_input.template)
        self.f = f
        input_segments = explained_input.input_segments
        group_assignments = util.standardize_groups(explained_input.group_assignments)
        mask_value = explained_input.mask_value
        merge_masks = explained_input.merge_masks

        # perturbed_inputs = None
        original_prediction = self.get_original_predicitons(input_segments)
        differences_by_scalarizer = []
        # For each scalarizer we may have a different ranking so need to create the perturbations for each scalarizer separately
        for scalarizer_index, attributions in enumerate(explained_input.attributions):
            if flip_signs:
                if (
                    original_prediction < 0.5
                ):  # If predict negative class, flip the signs of the attributions
                    attributions = -attributions
            rankings = util.get_group_ranking(attributions, is_descedning)
            input_perturbations = util.generate_perturbations(
                input_segments,
                group_assignments,
                rankings,
                mask_value,
                merge_masks,
                util.BitVectorSamplingType(direction.lower()),
                step_size,
                cut_off,
            )
            differences = self.compute_difference_original_to_perturbed(
                original_prediction, input_perturbations, scalarizer_index
            )
            differences_by_scalarizer.append(differences)

        return EvaluatedExplanation(
            input_segments,
            explained_input.template,
            group_assignments,
            explained_input.attributions,
            mask_value,
            merge_masks,
            direction,
            is_descedning,
            step_size,
            differences_by_scalarizer,
        )

    def visualize_evaluation(
        self,
        evaluated_explanations: list[EvaluatedExplanation],
        scalarizer_index: int,
        bin_size : float = 0.0,
        output_dir : str = None,
        std_display : str = "area",
    ):
        """
        Plot the graph of the differences in prediction at each percentage of the features perturbed averaged over all samples.
        If bin_size is None then only explanation may be provided and it will be plotted directly without binning or interpolation.
        If bin_size == 0.0, the graph is plotted using interpolation.
        If bin_size > 0.0, the graph is plotted with equal width binning.

        Parameters
        ----------
        evaluated_explanations : list[EvaluatedExplanation]
            The evaluated explanations to plot.
        scalarizer_index : int
            The index of the scalarizer for which the differences are to be plotted.
        bin_size : float
            The size of the bin to use for binning the results. If 0.0, interpolation is used.
        output_dir : str
            The directory to save the graph to.
        std_display : str
            The way to display the standard deviation. Default is None. 
            Options are "area" and "bar".
        """
        # Collect all the values needed from the evaluated explanations
        n = len(evaluated_explanations)
        direction = evaluated_explanations[0].direction.lower()
        direction_enum = util.BitVectorSamplingType(direction)
        value_at_zero = (
                0.0 if direction_enum == util.BitVectorSamplingType.DELETION else 1.0
            )
        is_descending = evaluated_explanations[0].is_descending
        differences_list = []
        percentages_by_step_list = [] # The percentage of features perturbed at each step for each evaluation
        for i in range(n):
            if evaluated_explanations[i].direction.lower() != direction:
                raise ValueError(
                    "All evaluated explanations should have the same direction of perturbation."+
                    f" Got {direction} for first evaluation, but {evaluated_explanations[i].direction} at {i}-th evaluation."
                )
            if evaluated_explanations[i].is_descending != is_descending:
                raise ValueError(
                    "All evaluated explanations should have the same is_descending value."+
                    f" Got {is_descending} for first evaluation, but {evaluated_explanations[i].is_descending} at {i}-th evaluation."
                )
            attributions = evaluated_explanations[i].attributions[scalarizer_index]
            differences = evaluated_explanations[i].differences[scalarizer_index]
            ranking = util.get_group_ranking(
                attributions, is_descending
            )
            percentage = util.get_size_percentages(evaluated_explanations[i].group_assignments)
            monotonic_percentage = util.get_monotonic_size_by_ranking(ranking, percentage)
            # Recreate the size of the perturbation at each step
            n_differences = len(differences)
            percentages_by_step = np.zeros(n_differences + 1)
            # If we do deletion then 0% perturbed means it's the original input, so the difference should be 0.0 and vice versa for insertion

            differences = np.insert(
                differences, 0, value_at_zero
            )
            step_size = evaluated_explanations[i].step_size
            cur_index = 1
            for j in range(step_size, n_differences,step_size,):
                percentages_by_step[cur_index] = sum(monotonic_percentage[j-step_size:j])
                cur_index += 1
            percentages_by_step[cur_index] = 1.0
            differences_list.append(differences)
            percentages_by_step_list.append(percentages_by_step)

        n_items_in_bin = None
        if bin_size is None:
            if len(evaluated_explanations) > 1:
                raise ValueError(
                    "Cannot plot graph directly when multiple evaluations are provided."
                )
            x_axis = percentages_by_step_list[0]
            y_axis = differences_list[0]
            std = None
        elif bin_size == 0.0:
            x_axis, y_axis, std = util.collect_with_interpolation(differences_list, percentages_by_step_list)
        else:
            n_bins = int(1.0 / bin_size) + 1
            x_axis, y_axis, std, n_items_in_bin = util.collect_results_in_bins(
                differences_list, percentages_by_step_list, bin_size=bin_size, n_bins=n_bins
            )

        plt.figure(figsize=(10, 6))
        if bin_size == 0.0:
            marker = None
        else:
            marker = "o"
        plt.plot(x_axis, y_axis, marker=marker, linestyle="-")
        y_max = max(y_axis) + 0.05
        if std is not None:
            y_max += + max(std) 
        plt.ylim(ymin=0, ymax=y_max)
        plt.xlim(xmin=0, xmax=1.01)
        plt.xticks(x_axis)
        if std is not None and std_display is not None:
            if std_display == "area":
                plt.fill_between(
                    x_axis,
                    np.array(y_axis) - np.array(std),
                    np.array(y_axis) + np.array(std),
                    alpha=0.2,
                )
            elif std_display == "bar":
                plt.errorbar(x_axis, y_axis, yerr=std, fmt="o")
            else:
                raise ValueError(f"Got unsupported std_display format {std_display}")

        if n_items_in_bin is not None:
            for i in range(len(n_items_in_bin)):
                plt.text(
                    x_axis[i],
                    y_axis[i] + 0.03,
                    f"{n_items_in_bin[i]}",
                    fontsize=10,
                    # horizontalalignment="right",
                )
        direction_name = direction_enum.name.lower()
        pretty_direction = "deleted" if direction_enum == util.BitVectorSamplingType.DELETION else "inserted"
        plt.xlabel(f"Percentage of {pretty_direction} features")
        plt.ylabel("Difference in prediction")
        plt.title(f"Difference in prediciton at each percentage of {pretty_direction} features")
        plt.grid(True, which="both")
        if output_dir is not None:
            os.makedirs(output_dir, exist_ok=True)
            if bin_size == 0.0:
                graph_name = f"{direction_name}_graph_n{n}.png"
            else:
                graph_name = f"{direction_name}_graph_bins{bin_size}_n{n}.png"
            plt.savefig(
                pjoin(
                    output_dir,
                    graph_name,
                )
            )
        plt.show()

    def edit_assignments(self,input : str, parser : str):
        """
        Given and input and a parser the function returns and edit widget that can be used to edit the input.

        params
        ------
        input: str
            The input as a string that will into assignable input segments.
        parser: str
            The parser to use to parse the input. Options are "word", "sentence", "syntax-tree".
        """
        try:
            from .demo.src.demo import EditWidget
        except ImportError:
            raise ImportError("The demo module is required to run this function. It is currently not part of the PyPI release and needs to be downloaded from the repository instead.")
        match parser.lower():
            case "word":
                input_segments = input.split()
                for i in range(len(input_segments) - 1):
                    input_segments[i] = input_segments[i] + " "
            case "sentence":
                input_segments = input.split(". ")
            case "syntax-tree":
                input_segments, _ = util.get_syntax_parse(input)
        self.edit_widget = EditWidget(inputSegments = input_segments)
        return self.edit_widget

    def get_edited_input(self):
        """
        Get the input after it has been edited using the edit_assignments function.
        """
        if self.edit_widget is None:
            raise ValueError("No edit widget has been created. Please run the edit_assignments function first.")
        input_segments = self.edit_widget.inputSegments
        assignments = self.edit_widget.assignments
        sample_name = self.edit_widget.sampleName
        direction = self.edit_widget.direction
        return AssignedInput(input_segments, assignments, direction, sample_name)

    def display_explanation(self, input_segments, group_assignments, attributions, sample_name = "Unnamed Sample"):
        """
        Display the explanation of the model's prediction on the input.

        params
        ------
        input_segments: list[str]
            The input segments of the input.
        group_assignments: list[int]
            The group assignments of the input segments.
        attributions: list[float]
            The attributions of the input segments.
        """
        try:
            from .demo.src.demo import DisplayWidget
        except:
            raise ImportError("The demo module is required to run this function. It is currently not part of the PyPI release and needs to be downloaded from the repository instead.")
        self.display_widget = DisplayWidget(
            inputSegments=input_segments,
            assignments=group_assignments,
            attributions=list(attributions),
            sampleName=sample_name,
        )
        return self.display_widget

    def get_masker_for_partitionshap(self, partition_tree):
        """Get the masker for the PartitionSHAP explainer."""

        def mask_to_input_wrapper(mask):
            input = self.mask_to_input(mask)
            return [input]

        # return self.GroupMasker(self.n_groups, mask_to_input_wrapper, partition_tree)

    # class GroupMasker(Masker):
    #     # The signatures of the methods here are based on the PartitionExplainer's Masker

    #     def __init__(self, n_groups, mask_fn, partition_tree):
    #         self.n_groups = n_groups
    #         self.mask_fn = mask_fn
    #         self.partition_tree = partition_tree

    #     def __call__(self, mask, x):
    #         return [mask]

    #     def shape(self, x):
    #         return (1, self.n_groups)

    #     def mask_shapes(self, s):
    #         return [((self.n_groups),)]

    #     def feature_names(self, s):
    #         """The names of the features for each mask position for the given input string."""
    #         return [[f"Group {i}" for i in range(self.n_groups)]]

    #     def clustering(self, *args):
    #         return self.partition_tree



import pickle

class AssignedInput: 
    
    def __init__(
        self,
        input_segments: list[str],
        group_assignments: list[int],
        direction = None,
        sample_name = None,
    ):
        self.input_segments = input_segments
        self.group_assignments = group_assignments
        self.direction = direction
        self.sample_name = sample_name

    def __str__(self):
        return f"Sample: {self.sample_name}\nWith input Segments: {self.input_segments}\nAssigned as: {self.group_assignments}\nTo be evaluated with: {self.direction}"
    def save(self, path: str):
        """Save the AssignedInput object to a file."""
        with open(path, "w") as f:
            pickle.dump(self, f)

    def load(path: str):
        """Load the AssignedInput object from a file."""
        with open(path, "r") as f:
            return pickle.load(f)

class ExplainedInput():

    def __init__(
        self,
        input_segments: list[str],
        template: str,
        group_assignments: list[int],
        attributions: list[list[float]],
        mask_value: str | list[str],
        merge_masks=False,
    ):
        self.input_segments = input_segments
        self.template = template
        self.group_assignments = group_assignments
        self.attributions = attributions
        self.mask_value = mask_value
        self.merge_masks = merge_masks

    def get_attributions(self):
        """A simple getter method that cleans up the case when the model output is one-dimensional."""
        if len(self.attributions) == 1:
            return self.attributions[0]
        else:
            return self.attributions

    def save(self, path: str):
        """Save the ExplainedInput object to a file."""
        with open(path, "w") as f:
            pickle.dump(self, f)

    def load(path: str):
        """Load the ExplainedInput object from a file."""
        with open(path, "r") as f:
            return pickle.load(f)


class EvaluatedExplanation(ExplainedInput):

    def __init__(
        self,
        input_segments: list[str],
        template: str,
        group_assignments: list[int],
        attributions: list[list[float]],
        mask_value: str | list[str],
        merge_masks: bool,
        direction: str,
        is_descending: bool,
        step_size: int,
        differences: list[list[float]],
    ):
        super().__init__(
            input_segments, template, group_assignments, attributions, mask_value, merge_masks
        )
        self.direction = direction
        self.is_descending = is_descending
        self.step_size = step_size
        self.differences = differences

    def save(self, path: str):
        """Save the EvaluatedExplanation object to a file."""
        with open(path, "w") as f:
            pickle.dump(self, f)

    def load(path: str):
        """Load the EvaluatedExplanation object from a file."""
        with open(path, "r") as f:
            return pickle.load(f)




from openai import OpenAI
import warnings
import pickle
import os
from os.path import join as pjoin
import hashlib
import dotenv


class ChatGPT:
    def __init__(
        self, n_responses=5, max_tokens=100, temperature=None, cache_dir=None
    ):
        """
        Parameters
        ----------
        n_responses : int
            The number of responses sampled for each input. Default is 5.
        max_tokens : int
            The maximum number of tokens to generate. Default is 100.
        cache_dir : str
            The directory to cache the responses. Default is None. If set to None responses will not be cached.
        """
        dotenv.load_dotenv()
        self.client = OpenAI()
        self.n_responses = n_responses
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.template: str = None
        self.cache_dir = cache_dir
        self.hasher = hashlib.sha256()
        if cache_dir:
            os.makedirs(cache_dir, exist_ok=True)
        self.call_index = 0

    def set_template(self, template: str):
        """Defines the template into which the input is inserted. The template should contain a placeholder for the input in the form {input}."""
        self.template = template
    
    def get_template(self):
        return self.template

    def __call__(
        self,
        input_string: str,
        n_responses: int = None,
        max_tokens: int = None,
    ):
        n_responses = n_responses if n_responses else self.n_responses
        max_tokens = max_tokens if max_tokens else self.max_tokens
        if not self.template:
            warnings.warn(
                "Template not set. Please set template before calling get_response. (Using empty template instead)"
            )
            self.template = "{input}"

        content = self.template.replace("{input}", input_string)
        self.hasher.update(content.encode())
        content_hash = self.hasher.hexdigest()
        self.hasher = hashlib.sha256()  # Reset the hasher

        if self.cache_dir is not None:
            cache_path = pjoin(self.cache_dir, f"{content_hash}.pkl")
            if os.path.exists(cache_path):
                completion = pickle.load(open(cache_path, "rb"))
        else:
            completion = self.client.chat.completions.create(
                model="gpt-4o-mini",
                messages=[
                    {
                        "role": "user",
                        "content": content,
                    },
                ],
                logprobs=False,
                n=n_responses,
                max_tokens=max_tokens,
                temperature=self.temperature,
            )
            if self.cache_dir:
                pickle.dump(completion, open(cache_path, "wb"))
        self.call_index += 1
        responses = []
        for choice in completion.choices:
            try:
                responses.append(choice.message.content.lower())
            except Exception as e:
                print("Got the following error while getting a response from the API:")
                print(e)
                print("Message in question: ", choice.message)
                print("Content in question: ", content)
        # print("Responses: ", responses)
        return responses



import numpy as np
import enum
import re
from nltk.tokenize import sent_tokenize

class BitVectorSamplingType(enum.Enum):
    DELETION = "deletion"
    INSERTION = "insertion"


def l1(a: float, b: float) -> float:
    return np.abs(a - b)


def l2(a: float, b: float) -> float:
    diff = a - b
    return np.sqrt(diff * diff)

justify_fns = {
    "START_WITH": lambda resp, value: resp.startswith(value),
    "END_WITH": lambda resp, value: resp.endswith(value),
    "EQUAL": lambda resp, value: resp == value,
    "CONTAIN": lambda resp, value: value in resp,
}
NLI_fns = ["SEMANTIC_EQUAL", "ENTAIL", "CONTRADICT"]


def evaluate_outcome(resps: list[str], operator : str, target : str, NLI_model=None):
    if type(resps) is str:
        resps = [resps]

    value = target.lower()

    if operator in justify_fns:
        justify_fn = justify_fns[operator]
        justifications = [justify_fn(resp, value) for resp in resps]
        return sum(justifications) / len(justifications)
    elif operator in NLI_fns:
        if NLI_model is None:
            raise ValueError(f"Trying to run NLI method: {operator} but no NLI model is defined")
        NLI_predictions = []
        for resp in resps:
            prediction = NLI_model([dict(text=resp, text_pair=value)], top_k=None)
            if operator == "ENTAIL":
                for labelled_score in prediction[0]:
                    if labelled_score["label"] == "entailment":
                        NLI_predictions.append(labelled_score["score"])
            elif operator == "CONTRADICT":
                for labelled_score in prediction[0]:
                    if labelled_score["label"] == "contradiction":
                        NLI_predictions.append(labelled_score["score"])
            elif operator == "SEMANTIC_EQUAL":
                predction_backward = NLI_model([dict(text=value, text_pair=resp)], top_k=None)
                foward_score = 0.0
                for labelled_score in prediction[0]:
                    if labelled_score["label"] == "entailment":
                        foward_score = labelled_score["score"]
                for labelled_score in predction_backward[0]:
                    if labelled_score["label"] == "entailment":
                        backward_score = labelled_score["score"]
                aggregate = min([foward_score, backward_score])
                NLI_predictions.append(aggregate)
        return sum(NLI_predictions) / len(NLI_predictions)
    else:
        raise ValueError(f"Operator {operator} not found")


def word_tokenize(passage : str):
    """
    A simplistic word tokenizer that splits the passage by whitespace.
    Importantly it retains the white spaces as part of the segments allowing the sequence to be reconstructed. 
    """
    re_S = re.compile(r"(\S+)")
    segments_with_white_space = re_S.split(passage)
    segments = []
    for j in range(1, len(segments_with_white_space) - 1, 2):
        segments.append(segments_with_white_space[j] + segments_with_white_space[j + 1])
    return segments


def sentence_tokenize(passage : str):
    """
    Tokenize the passage into sentences by using nltk and then trying to recover the white space afterwards. 
    """
    sents = sent_tokenize(passage)
    shredded_passage = passage
    for i,sent in enumerate(sents):
        shredded_passage = shredded_passage[len(sent):]
        if len(shredded_passage) == 0: 
            break
        begin_non_white_space = re.match(r"\s*", shredded_passage, re.UNICODE).end()
        sents[i] += shredded_passage[:begin_non_white_space]
        shredded_passage = shredded_passage[begin_non_white_space:]

    return sents


def paragraph_tokenize(passage : str):
    """
    Tokenize the passage into paragraphs where the end of the paragraph is defined by at least two new lines.
    """
    passage = re.sub("\n\n+", "\n\n", passage)
    paragraphs = passage.split("\n\n")
    for i in range(len(paragraphs) - 1):
        paragraphs[i] = paragraphs[i] + "\n\n"
    return paragraphs


def syntax_tokenize(input_string):
    raise NotImplementedError("This function is currenlty being worked on.")
    from colab.parser.parser import parse_passage

    base_parse = parse_passage(input_string)
    groups = []
    for parse_group in base_parse["nodes"]:
        ids = []
        children = parse_group["children"]
        if len(children) != 0:
            for child in children:
                if child["type"] == "span":
                    ids.append(child["id"])
        else:
            ids = [parse_group["data"]["spans"][0]]
        groups.append(ids)
    colab_atomics = [span["text"] for span in base_parse["spans"]]
    n_values = len(colab_atomics)
    assignments = -1 * np.ones(n_values, dtype=int)
    for i, group in enumerate(groups):
        for value in group:
            assignments[value] = i
    straggler_index = len(groups)
    for i, assignment in enumerate(assignments):
        if assignment == -1:
            assignments[i] = straggler_index
            straggler_index += 1
    return colab_atomics, assignments


def standardize_groups(group_assignments: list[int]) -> list[int]:
    """Standardize the groups assignments to be 0-indexed and contiguous."""
    groups = set(group_assignments)
    n_groups = len(groups)
    groups_dict = dict(zip(groups, range(n_groups)))
    return [groups_dict[c] for c in group_assignments]


def get_size_percentages(group_assignments):
        """Get the size of each group as a percentage of the total number of features."""
        n_features = len(group_assignments)
        n_groups = len(set(group_assignments))
        n_features_by_group = np.zeros(n_groups)
        for i in range(n_features):
            n_features_by_group[group_assignments[i]] += 1
        return n_features_by_group / n_features

def get_group_ranking(attributions : list[float], is_descedning=True):
    """Define the rankings of the groups for each input based on the attribution scores. This is only done conditionally
    as directions like Random do not require this."""

    if is_descedning:
        # Reverse the order to go from highest value to lowest
        return np.argsort(attributions)[::-1]
    else:
        return np.argsort(attributions)

def get_monotonic_size_by_ranking(ranking : list[int], sizes : list[float]):
    """Given a size (percentage or number of features) of each group, return the summed sizes in a monotonic order according to self.ranking."""
    monotonic_sizes = np.zeros(len(sizes))
    prev_size = 0
    for i in range(len(sizes)):
        monotonic_sizes[i] = prev_size + sizes[ranking[i]]
        prev_size = monotonic_sizes[i]
    return monotonic_sizes

def get_number_of_constituents_by_group(group_assignments : list[int]):
    """Get the number of constituents by group."""
    n_constituents = np.zeros(
        len(set(group_assignments)), dtype=np.int32
    )  
    for i in range(len(group_assignments)):
        n_constituents[group_assignments[i]] += 1
    return n_constituents

def mask_input_by_allowed_groups_string(input_segments, group_assignments, allowed_groups, mask_value, merge_masks):
    """Mask the input at the given index by masking all the features that are not in the allowed groups."""
    result_string = ""
    prev_was_masked = False
    for i in range(len(input_segments)):
        if group_assignments[i] in allowed_groups:
            result_string += input_segments[
                i
            ]  # Assumes necessary spaces are included in the input parts
            prev_was_masked = False
        else:
            # lg.debug(f"masking: {self.inputs[input_index][i]}")
            if merge_masks and prev_was_masked:
                continue
            result_string += mask_value
            prev_was_masked = True
    return result_string

# def mask_input_by_allowed_groups_tokens(self, input_index, allowed_groups):
#     input = []
#     prev_was_masked = False
#     for i in range(self.input_lengths[input_index]):
#         # If mask is False, then the input is replaced with the mask value for that input
#         # Else the input is included as is
#         # fmt: off
#         if not self.group_assignments[input_index][i] in allowed_groups:
#             if not prev_was_masked or (not self.merge_masks):
#                 # Place the mask value if 1. not merging masks or 2. merging masks but the previous input was not masked
#                 if self.mask_value is not None: # Use mask value None to drop tokens completely
#                     input.append(self.mask_value)
#             prev_was_masked = True
#         else:
#             input.append(self.inputs[input_index][i])
#             prev_was_masked = False
#     # lg.debug(f"Input generated from mask: {input}")
#     if len(input) == 1:
#         return input[0].unsqueeze(0)
#     return torch.stack(input)

def generate_perturbations(
        input_segments : list[str],
        group_assignments : list[int],
        ranking : list[int],
        mask_value: str, 
        merge_masks: bool, 
        direction : BitVectorSamplingType, 
        step_size : int, 
        cut_off : float
    ):
    """Generate the perturbed inputs for the insertion direction."""
    perturbations = []
    allowed_groups = []
    cur_index = step_size  # Begin with first step_size many groups
    percentages = get_size_percentages(group_assignments)
    percent_perturbed = 0  # If deletion this is number of deleted features, if insertion this is number of inserted features
    while percent_perturbed < cut_off - np.finfo(float).eps:
        if direction == BitVectorSamplingType.INSERTION:
            # Only allow the groups up to cur_index (inserting the first cur_index many groups)
            allowed_groups = ranking[:cur_index]
            percent_perturbed = sum(percentages[allowed_groups])

        elif direction == BitVectorSamplingType.DELETION:
            # Only allow the groups past the current index (deleting the first cur_index many groups)
            allowed_groups = ranking[cur_index:]
            percent_perturbed = 1 - sum(percentages[allowed_groups])
        else:
            raise NotImplementedError(
                "Only deletion and insertion directions are implemented."
            )
        perturbations.append(
            mask_input_by_allowed_groups_string(input_segments,group_assignments, allowed_groups, mask_value, merge_masks)
        )
        cur_index += step_size

    return perturbations

def compute_area(differences : list[float], group_assignments : list[int], ranking : list[int]):
    from sklearn.metrics import auc
    """Compute the area as the difference value times the number of constituents of the group or the percentage of the group."""
    
    percentages = get_size_percentages(group_assignments) 
    monotonic_percentages_by_input = get_monotonic_size_by_ranking(ranking, percentages)
    return auc(monotonic_percentages_by_input, differences)

def collect_results_directly(self, percentages):
    """Collect the results directly without binning."""
    percentage_to_diff = {}  # maps percentage -> [sum of differences, number of differences]
    for i, differences in enumerate(self.diff):
        for j in range(len(differences)):
            percentage = sum([percentages[i][c] for c in self.allowed_groups_by_input[i][j]])
            if self.direction == BitVectorSamplingType.DELETION:
                percentage = 1.0 - percentage
            if percentage not in percentage_to_diff:
                percentage_to_diff[percentage] = [0.0, 0]
            percentage_to_diff[percentage][0] += differences[j]
            percentage_to_diff[percentage][1] += 1
    percentages = sorted(percentage_to_diff.keys())
    avg_diffs = [percentage_to_diff[p][0] / percentage_to_diff[p][1] for p in percentages]
    items_in_bin = [percentage_to_diff[p][1] for p in percentages]
    return percentages, avg_diffs, items_in_bin


def collect_results_in_bins(
    differences_list, percentages_by_step_list, bin_size=0.05, n_bins=10
):
    """Collect the results in bins of size bin_size in n_bins."""
    bins = np.zeros(n_bins)
    n_items_in_bin = np.zeros(n_bins, dtype=int)
    values_in_bin = {i: [] for i in range(n_bins)}  # To calculate the standard deviation
    for i, differences in enumerate(differences_list):
        for j in range(len(differences)):
            percentage = percentages_by_step_list[i][j]
            bin_index = int(percentage / bin_size)
            if bin_index >= n_bins:
                bin_index = n_bins - 1
            bins[bin_index] += differences[j]
            n_items_in_bin[bin_index] += 1
            values_in_bin[bin_index].append(differences[j])
    avg_diffs = bins / n_items_in_bin
    used_bins = []
    used_bin_values = []
    used_bin_values_std = []
    non_zero_n_items = []
    for i in range(n_bins):
        if n_items_in_bin[i] > 0:
            if i < n_bins - 1:
                # Display the bin's x-position as the middle of the bin
                used_bins.append((i + 0.5) * bin_size)
            else:
                used_bins.append(i * bin_size)
            used_bin_values.append(avg_diffs[i])
            var_accumalator = 0
            for value in values_in_bin[i]:
                var_accumalator += (value - avg_diffs[i]) ** 2
            used_bin_values_std.append(np.sqrt(var_accumalator / n_items_in_bin[i]))
            non_zero_n_items.append(n_items_in_bin[i])
    return used_bins, used_bin_values, used_bin_values_std, non_zero_n_items


def collect_with_interpolation(differences_list, percentages_by_step_list):
    x_axes = percentages_by_step_list
    y_axes = differences_list
    mean_x_axis = [i * 0.05 for i in range(21)]
    ys_interp = [np.interp(mean_x_axis, x_axes[i], y_axes[i]) for i in range(len(x_axes))]
    mean_y_axis = np.mean(ys_interp, axis=0)
    std_y_axis = np.std(ys_interp, axis=0)
    return mean_x_axis, mean_y_axis, std_y_axis
